# Copyright 2018 Comcast Cable Communications Management, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TODO: Add module docstring."""

import logging
from datetime import datetime
from hashlib import sha256

import requests.compat as urlparse
from boto.dynamodb2.layer1 import DynamoDBConnection

logger = logging.getLogger(__name__)

__all__ = [u'BotoRequestSigner']


class BotoRequestSigner(object):
    """TODO: Add class docstring."""

    def __init__(self, index_url, access_key, secret_access_key):
        """TODO: Add method docstring."""
        url = urlparse.urlparse(index_url)
        self.boto_connection = DynamoDBConnection(
            host=url.hostname,
            port=url.port,
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_access_key,
            is_secure=False)

    @staticmethod
    def __canonical_date(headers):
        """
        Derive canonical date (ISO 8601 string).

        Either from headers (if possible) or synthesize it if no usable header exists.
        """
        iso_format = u'%Y%m%dT%H%M%SZ'
        http_format = u'%a, %d %b %Y %H:%M:%S GMT'

        def try_parse(date_string, format):
            if date_string is None:
                return None
            try:
                return datetime.strptime(date_string, format)
            except ValueError:
                return None

        amz_date = try_parse(headers.get(u'X-Amz-Date'), iso_format)
        http_date = try_parse(headers.get(u'Date'), http_format)
        fallback_date = datetime.utcnow()

        date = next(d for d in [amz_date, http_date, fallback_date] if d is not None)
        return date.strftime(iso_format)

    def build_auth_header(self, method, path, headers, body, params=None):
        """Construct an Authorization header, using boto."""
        request = self.boto_connection.build_base_http_request(
            method=method,
            path=path,
            auth_path=path,
            headers=headers,
            data=body,
            params=params or {})

        auth_handler = self.boto_connection._auth_handler

        timestamp = BotoRequestSigner.__canonical_date(headers)
        request.timestamp = timestamp[0:8]

        request.region_name = u'us-east-1'
        request.service_name = u'VinylDNS'

        credential_scope = u'/'.join([request.timestamp, request.region_name, request.service_name, u'aws4_request'])

        canonical_request = auth_handler.canonical_request(request)
        split_request = canonical_request.split('\n')

        if params != {} and split_request[2] == '':
            split_request[2] = generate_canonical_query_string(params)
            canonical_request = '\n'.join(split_request)

        hashed_request = sha256(canonical_request.encode(u'utf-8')).hexdigest()

        string_to_sign = u'\n'.join([u'AWS4-HMAC-SHA256', timestamp, credential_scope, hashed_request])
        signature = auth_handler.signature(request, string_to_sign)
        headers_to_sign = auth_handler.headers_to_sign(request)

        auth_header = u','.join([
            u'AWS4-HMAC-SHA256 Credential=%s' % auth_handler.scope(request),
            u'SignedHeaders=%s' % auth_handler.signed_headers(headers_to_sign),
            u'Signature=%s' % signature])

        return auth_header


def generate_canonical_query_string(params):
    """
    Using in place of canonical_query_string from boto/auth.py to support POST requests with query parameters
    """
    post_params = []
    for param in sorted(params):
        value = params[param].encode('utf-8')
        import urllib
        post_params.append('%s=%s' % (urllib.parse.quote(param, safe='-_.~'),
                           urllib.parse.quote(value, safe='-_.~')))
    return '&'.join(post_params)



# Copyright 2018 Comcast Cable Communications Management, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TODO: Add module docstring."""

from vinyldns.zone import Zone

from vinyldns.serdes import parse_datetime, map_option


class RecordType:
    A = "A"
    AAAA = "AAAA"
    CNAME = "CNAME"
    PTR = "PTR"
    MX = "MX"
    NS = "NS"
    SOA = "SOA"
    SRV = "SRV"
    TXT = "TXT"
    SSHFP = "SSHFP"
    SPF = "SPF"
    UNKNOWN = "UNKNOWN"


class RecordSetStatus:
    Active = "Active"
    Inactive = "Inactive"
    Pending = "Pending"
    PendingUpdate = "PendingUpdate"
    PendingDelete = "PendingDelete"


class AData(object):
    def __init__(self, address):
        self.address = address

    @staticmethod
    def from_dict(d):
        return AData(d['address'])


class AAAAData(object):
    def __init__(self, address):
        self.address = address

    @staticmethod
    def from_dict(d):
        return AAAAData(d['address'])


class CNAMEData(object):
    def __init__(self, cname):
        self.cname = cname

    @staticmethod
    def from_dict(d):
        return CNAMEData(d['cname'])


class MXData(object):
    def __init__(self, preference, exchange):
        self.preference = preference
        self.exchange = exchange

    @staticmethod
    def from_dict(d):
        return MXData(d['preference'], d['exchange'])


class NSData(object):
    def __init__(self, nsdname):
        self.nsdname = nsdname

    @staticmethod
    def from_dict(d):
        return NSData(d['nsdname'])


class PTRData(object):
    def __init__(self, ptrdname):
        self.ptrdname = ptrdname

    @staticmethod
    def from_dict(d):
        return PTRData(d['ptrdname'])


class SOAData(object):
    def __init__(self, mname, rname, serial, refresh, retry, expire, minimum):
        self.mname = mname
        self.rname = rname
        self.serial = serial
        self.refresh = refresh
        self.retry = retry
        self.expire = expire
        self.minimum = minimum

    @staticmethod
    def from_dict(d):
        return SOAData(d['mname'], d['rname'], d['serial'], d['refresh'], d['retry'], d['expire'], d['minimum'])


class SPFData(object):
    def __init__(self, text):
        self.text = text

    @staticmethod
    def from_dict(d):
        return SPFData(d['text'])


class SRVData(object):
    def __init__(self, priority, weight, port, target):
        self.priority = priority
        self.weight = weight
        self.port = port
        self.target = target

    @staticmethod
    def from_dict(d):
        return SRVData(d['priority'], d['weight'], d['port'], d['target'])


class SSHFPData(object):
    def __init__(self, algorithm, type, fingerprint):
        self.algorithm = algorithm
        self.type = type
        self.fingerprint = fingerprint

    @staticmethod
    def from_dict(d):
        return SSHFPData(d['algorithm'], d['type'], d['fingerprint'])


class TXTData(object):
    def __init__(self, text):
        self.text = text

    @staticmethod
    def from_dict(d):
        return TXTData(d['text'])


class UNKNOWNData(object):
    def __init__(self, rdata):
        self.rdata = rdata

    @staticmethod
    def from_dict(d):
        return UNKNOWNData(d['rdata'])


# A mapping of record types to functions that create an RData instance from a dictionary
rdata_converters = {
    RecordType.A: AData.from_dict,
    RecordType.AAAA: AAAAData.from_dict,
    RecordType.CNAME: CNAMEData.from_dict,
    RecordType.PTR: PTRData.from_dict,
    RecordType.MX: MXData.from_dict,
    RecordType.NS: NSData.from_dict,
    RecordType.SOA: SOAData.from_dict,
    RecordType.SRV: SRVData.from_dict,
    RecordType.TXT: TXTData.from_dict,
    RecordType.SSHFP: SSHFPData.from_dict,
    RecordType.SPF: SPFData.from_dict,
    RecordType.UNKNOWN: UNKNOWNData.from_dict
}


class RecordSet(object):
    def __init__(self, zone_id, name, type, ttl, status=None, created=None,
                 updated=None, records=[], id=None, owner_group_id=None):
        self.zone_id = zone_id
        self.name = name
        self.type = type
        self.ttl = ttl
        self.status = status
        self.created = created
        self.updated = updated
        self.records = records
        self.id = id
        self.owner_group_id = owner_group_id

    @staticmethod
    def from_dict(d):
        return RecordSet(
            zone_id=d['zoneId'],
            name=d['name'],
            type=d['type'],
            ttl=d['ttl'],
            status=d.get('status'),
            created=map_option(d.get('created'), parse_datetime),
            updated=d.get('updated'),
            records=[rdata_converters[d['type']](rd) for rd in d.get('records', [])],
            id=d.get('id'),
            owner_group_id=d.get('ownerGroupId')
        )


class ListRecordSetsResponse(object):
    def __init__(self, record_sets, start_from=None, next_id=None, max_items=None, record_name_filter=None):
        self.record_sets = record_sets
        self.start_from = start_from
        self.next_id = next_id
        self.max_items = max_items
        self.record_name_filter = record_name_filter

    @staticmethod
    def from_dict(d):
        return ListRecordSetsResponse(
            record_sets=[RecordSet.from_dict(elem) for elem in d.get('recordSets', [])],
            start_from=d.get('startFrom'),
            next_id=d.get('nextId'),
            max_items=d.get('maxItems'),
            record_name_filter=d.get('recordNameFilter')
        )


class RecordSetChange(object):
    def __init__(self, zone, record_set, user_id, change_type, status, created, system_message, updates, id,
                 user_name):
        self.zone = zone
        self.record_set = record_set
        self.user_id = user_id
        self.change_type = change_type
        self.status = status
        self.created = created
        self.system_message = system_message
        self.updates = updates
        self.id = id
        self.user_name = user_name

    @staticmethod
    def from_dict(d):
        return RecordSetChange(
            zone=Zone.from_dict(d['zone']),
            record_set=RecordSet.from_dict(d['recordSet']),
            user_id=d['userId'],
            change_type=d['changeType'],
            status=d['status'],
            created=map_option(d.get('created'), parse_datetime),
            system_message=d.get('systemMessage'),
            updates=map_option(d.get('updates'), RecordSet.from_dict),
            id=d['id'],
            user_name=d.get('userName')
        )


class ListRecordSetChangesResponse(object):
    def __init__(self, zone_id, record_set_changes, start_from=None, next_id=None, max_items=100):
        self.zone_id = zone_id
        self.record_set_changes = record_set_changes
        self.next_id = next_id
        self.start_from = start_from
        self.max_items = max_items

    @staticmethod
    def from_dict(d):
        changes = [RecordSetChange.from_dict(elem) for elem in d.get('recordSetChanges', [])]
        return ListRecordSetChangesResponse(zone_id=d['zoneId'], record_set_changes=changes,
                                            next_id=d.get('nextId'), start_from=d.get('startFrom'),
                                            max_items=d['maxItems'])



# Copyright 2018 Comcast Cable Communications Management, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TODO: Add module docstring."""

from vinyldns.serdes import parse_datetime, map_option, to_utc_strftime
from vinyldns.record import rdata_converters
import json


def to_review_json(s):
    """
    Converts the string to json
    :param s: A string
    :return: A json formatted string representation of the object
    """
    b = {}
    if s:
        b.update({"reviewComment": s})
        return json.dumps(b)
    else:
        return None


class AddRecord(object):
    def __init__(self, input_name, type, ttl, record):
        self.input_name = input_name
        self.type = type
        self.ttl = ttl
        self.record = record
        self.change_type = 'Add'

    @staticmethod
    def from_dict(d):
        return AddRecord(
            input_name=d['inputName'],
            type=d['type'],
            ttl=d['ttl'],
            record=d['record']
        )


class DeleteRecordSet(object):
    def __init__(self, input_name, type, record=None):
        self.input_name = input_name
        self.type = type
        self.record = record
        self.change_type = 'DeleteRecordSet'

    @staticmethod
    def from_dict(d):
        return DeleteRecordSet(
            input_name=d['inputName'],
            type=d['type'],
            record=map_option(d.get('record'), rdata_converters[d['type']])
        )


class ValidationError(object):
    def __init__(self, error_type, message):
        self.error_type = error_type
        self.message = message

    @staticmethod
    def from_dict(d):
        return ValidationError(
            error_type=d['errorType'],
            message=d['message']
        )


class BatchChangeRequest(object):
    change_type_converters = {
        'Add': AddRecord.from_dict,
        'DeleteRecordSet': DeleteRecordSet.from_dict
    }

    def __init__(self, changes, comments=None, owner_group_id=None,
                 scheduled_time=None):
        self.comments = comments
        self.changes = changes
        self.owner_group_id = owner_group_id
        self.scheduled_time = to_utc_strftime(scheduled_time) if scheduled_time else None

    @staticmethod
    def from_dict(d):
        return BatchChangeRequest(
            comments=d.get('comments'),
            changes=[BatchChangeRequest.change_type_converters[elem['changeType']](elem)
                     for elem in d.get('changes', [])],
            owner_group_id=d.get('ownerGroupId'),
            scheduled_time=map_option(d.get('scheduledTime'), parse_datetime)
        )


class AddRecordChange(object):
    def __init__(self, zone_id, zone_name, record_name, input_name, type, ttl,
                 record, status, id, validation_errors, system_message=None,
                 record_change_id=None, record_set_id=None):
        self.zone_id = zone_id
        self.zone_name = zone_name
        self.record_name = record_name
        self.input_name = input_name
        self.type = type
        self.ttl = ttl
        self.record = record
        self.status = status
        self.id = id
        self.system_message = system_message
        self.record_change_id = record_change_id
        self.record_set_id = record_set_id
        self.change_type = 'Add'
        self.validation_errors = validation_errors

    @staticmethod
    def from_dict(d):
        return AddRecordChange(
            zone_id=d['zoneId'],
            zone_name=d['zoneName'],
            record_name=d['recordName'],
            input_name=d['inputName'],
            type=d['type'],
            ttl=d['ttl'],
            record=rdata_converters[d['type']](d['record']),
            status=d['status'],
            id=d['id'],
            system_message=d.get('systemMessage'),
            record_change_id=d.get('recordChangeId'),
            record_set_id=d.get('recordSetId'),
            validation_errors=[ValidationError.from_dict(elem) for elem in d.get('validationErrors', [])]
        )


class DeleteRecordSetChange(object):
    def __init__(self, zone_id, zone_name, record_name, input_name, type, status,
                 id, validation_errors, system_message=None, record_change_id=None, record_set_id=None, record=None):
        self.zone_id = zone_id
        self.zone_name = zone_name
        self.record_name = record_name
        self.input_name = input_name
        self.type = type
        self.record = record
        self.status = status
        self.id = id
        self.system_message = system_message
        self.record_change_id = record_change_id
        self.record_set_id = record_set_id
        self.change_type = 'DeleteRecordSet'
        self.validation_errors = validation_errors

    @staticmethod
    def from_dict(d):
        return DeleteRecordSetChange(
            zone_id=d['zoneId'],
            zone_name=d['zoneName'],
            record_name=d['recordName'],
            input_name=d['inputName'],
            type=d['type'],
            record=map_option(d.get('record'), rdata_converters[d['type']]),
            status=d['status'],
            id=d['id'],
            system_message=d.get('systemMessage'),
            record_change_id=d.get('recordChangeId'),
            record_set_id=d.get('recordSetId'),
            validation_errors=[ValidationError.from_dict(elem) for elem in d.get('validationErrors', [])]
        )


class BatchChange(object):
    change_type_converters = {
        'Add': AddRecordChange.from_dict,
        'DeleteRecordSet': DeleteRecordSetChange.from_dict
    }

    def __init__(self, user_id, user_name, created_timestamp, changes, id,
                 status, approval_status, comments=None, owner_group_id=None,
                 owner_group_name=None, reviewer_id=None,
                 reviewer_user_name=None, review_comment=None,
                 review_timestamp=None, scheduled_time=None):
        self.user_id = user_id
        self.user_name = user_name
        self.comments = comments
        self.created_timestamp = created_timestamp
        self.changes = changes
        self.id = id
        self.status = status
        self.owner_group_id = owner_group_id
        self.owner_group_name = owner_group_name
        self.approval_status = approval_status
        self.reviewer_id = reviewer_id
        self.reviewer_user_name = reviewer_user_name
        self.review_comment = review_comment
        self.review_timestamp = review_timestamp
        self.scheduled_time = scheduled_time

    @staticmethod
    def from_dict(d):
        return BatchChange(
            user_id=d['userId'],
            user_name=d['userName'],
            comments=d.get('comments'),
            created_timestamp=map_option(d.get('createdTimestamp'), parse_datetime),
            changes=[BatchChange.change_type_converters[elem['changeType']](elem) for elem in d.get('changes', [])],
            id=d['id'],
            status=d['status'],
            owner_group_id=d.get('ownerGroupId'),
            owner_group_name=d.get('ownerGroupName'),
            approval_status=d['approvalStatus'],
            reviewer_id=d.get('reviewerId'),
            reviewer_user_name=d.get('reviewerUserName'),
            review_comment=d.get('reviewComment'),
            review_timestamp=map_option(d.get('reviewTimestamp'), parse_datetime),
            scheduled_time=map_option(d.get('scheduledTime'), parse_datetime)
        )


class BatchChangeSummary(object):
    def __init__(self, user_id, user_name, created_timestamp, total_changes, id,
                 status, approval_status, comments=None, owner_group_id=None,
                 owner_group_name=None, reviewer_id=None,
                 reviewer_user_name=None, review_comment=None,
                 review_timestamp=None, scheduled_time=None):
        self.user_id = user_id
        self.user_name = user_name
        self.comments = comments
        self.created_timestamp = created_timestamp
        self.total_changes = total_changes
        self.status = status
        self.id = id
        self.owner_group_id = owner_group_id
        self.owner_group_name = owner_group_name
        self.approval_status = approval_status
        self.reviewer_id = reviewer_id
        self.reviewer_user_name = reviewer_user_name
        self.review_comment = review_comment
        self.review_timestamp = review_timestamp
        self.scheduled_time = scheduled_time

    @staticmethod
    def from_dict(d):
        return BatchChangeSummary(
            user_id=d['userId'],
            user_name=d['userName'],
            comments=d.get('comments'),
            created_timestamp=parse_datetime(d['createdTimestamp']),
            total_changes=d['totalChanges'],
            status=d['status'],
            id=d['id'],
            owner_group_id=d.get('ownerGroupId'),
            owner_group_name=d.get('ownerGroupName'),
            approval_status=d['approvalStatus'],
            reviewer_id=d.get('reviewerId'),
            reviewer_user_name=d.get('reviewerUserName'),
            review_comment=d.get('reviewComment'),
            review_timestamp=d.get('reviewTimestamp'),
            scheduled_time=map_option(d.get('scheduledTime'), parse_datetime)
        )


class ListBatchChangeSummaries(object):
    def __init__(self, batch_changes, start_from=None, next_id=None,
                 max_items=100, ignore_access=False, approval_status=None):
        self.batch_changes = batch_changes
        self.start_from = start_from
        self.next_id = next_id
        self.max_items = max_items
        self.ignore_access = ignore_access
        self.approval_status = approval_status

    @staticmethod
    def from_dict(d):
        return ListBatchChangeSummaries(
            batch_changes=[BatchChangeSummary.from_dict(elem) for elem in d.get('batchChanges', [])],
            start_from=d.get('startFrom'),
            next_id=d.get('nextId'),
            max_items=d.get('maxItems', 100),
            ignore_access=d.get('ignoreAccess'),
            approval_status=d.get('approvalStatus')
        )



# Copyright 2018 Comcast Cable Communications Management, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TODO: Add module docstring."""

from vinyldns.serdes import map_option, parse_datetime


class UserLockStatus:
    Unlocked = "Unlocked"
    Locked = "Locked"


class User(object):
    def __init__(self, id, user_name=None, first_name=None, last_name=None, email=None, created=None,
                 lock_status=UserLockStatus.Unlocked):
        self.id = id
        self.user_name = user_name
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.created = created
        self.lock_status = lock_status

    @staticmethod
    def from_dict(d):
        created = map_option(d.get('created'), parse_datetime)
        return User(
            id=d['id'],
            user_name=d.get('userName'),
            first_name=d.get('firstName'),
            last_name=d.get('lastName'),
            email=d.get('email'),
            created=created,
            lock_status=d.get('lockStatus', UserLockStatus.Unlocked)
        )


class Group(object):
    def __init__(self, name, email, description=None, created=None, members=[], admins=[], id=None):
        self.name = name
        self.email = email
        self.description = description
        self.created = created
        self.members = members
        self.admins = admins
        self.id = id

    @staticmethod
    def from_dict(d):
        return Group(
            name=d['name'],
            email=d['email'],
            description=d.get('description'),
            created=map_option(d.get('created'), parse_datetime),
            members=[User.from_dict(ud) for ud in d.get('members', [])],
            admins=[User.from_dict(ud) for ud in d.get('admins', [])],
            id=d.get('id')
        )


class ListGroupsResponse(object):
    def __init__(self, groups, max_items=None, group_name_filter=None, start_from=None, next_id=None):
        self.groups = groups
        self.max_items = max_items
        self.group_name_filter = group_name_filter
        self.start_from = start_from
        self.next_id = next_id

    @staticmethod
    def from_dict(d):
        return ListGroupsResponse(

            groups=[Group.from_dict(elem) for elem in d.get('groups', [])],
            max_items=d.get('maxItems'),
            group_name_filter=d.get('groupNameFilter'),
            start_from=d.get('startFrom'),
            next_id=d.get('nextId')
        )


class GroupChange(object):
    def __init__(self, new_group, change_type, user_id, old_group, id, created):
        self.new_group = new_group
        self.change_type = change_type
        self.user_id = user_id
        self.old_group = old_group
        self.id = id
        self.created = created

    @staticmethod
    def from_dict(d):
        return GroupChange(
            new_group=Group.from_dict(d['newGroup']),
            change_type=d['changeType'],
            user_id=d['userId'],
            old_group=map_option(d.get('oldGroup'), Group.from_dict),
            id=d['id'],
            created=map_option(d.get('created'), parse_datetime)
        )


class ListGroupChangesResponse(object):
    def __init__(self, changes, start_from=None, next_id=None, max_items=100):
        self.changes = changes
        self.start_from = start_from
        self.next_id = next_id
        self.max_items = max_items

    @staticmethod
    def from_dict(d):
        return ListGroupChangesResponse(
            changes=[GroupChange.from_dict(elem) for elem in d.get('changes', [])],
            start_from=d.get('startFrom'),
            next_id=d.get('nextId'),
            max_items=d.get('maxItems')
        )


class Member(object):
    def __init__(self, id, user_name=None, first_name=None, last_name=None, email=None, created=None, is_admin=False):
        self.id = id
        self.user_name = user_name
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.created = created
        self.is_admin = is_admin

    @staticmethod
    def from_dict(d):
        return Member(
            id=d['id'],
            user_name=d.get('userName'),
            first_name=d.get('firstName'),
            last_name=d.get('lastName'),
            email=d.get('email'),
            created=map_option(d.get('created'), parse_datetime),
            is_admin=d.get('isAdmin', False)
        )


class ListMembersResponse(object):
    def __init__(self, members, start_from=None, next_id=None, max_items=100):
        self.members = members
        self.start_from = start_from
        self.next_id = next_id
        self.max_items = max_items

    @staticmethod
    def from_dict(d):
        return ListMembersResponse(
            members=[Member.from_dict(elem) for elem in d.get('members', [])],
            start_from=d.get('startFrom'),
            next_id=d.get('nextId'),
            max_items=d['maxItems']
        )


class ListAdminsResponse(object):
    def __init__(self, admins):
        self.admins = admins

    @staticmethod
    def from_dict(d):
        return ListAdminsResponse(
            admins=[User.from_dict(elem) for elem in d.get('admins', [])]
        )



# Copyright 2018 Comcast Cable Communications Management, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TODO: Add module docstring."""
import re
from datetime import date, datetime
import json
# Python 2/3 compatibility
try:
    from datetime import timezone
except ImportError:
    from dateutil.tz import tzutc

camel_pat = re.compile(r'([A-Z])')
under_pat = re.compile(r'_([a-z])')


def camel_to_underscore(name):
    """
    Given a string in camelCase, convert to snake_case
    :param name: The string to be converted
    :return: A String formatted in snake_case
    """
    return camel_pat.sub(lambda x: '_' + x.group(1).lower(), name)


def underscore_to_camel(name):
    """
    Given a string in snake_case, convert to camelCase
    :param name: The string to be converted
    :return: A String formatted in camelCase
    """
    return under_pat.sub(lambda x: x.group(1).upper(), name)


def to_dict(obj, cls=None):
    """
    Given an object, serializes all fields to a dictionary including nested objects.

    Converts the field names to camelCase along the way suitable for convention in the API
    :param obj: An object instance
    :param cls: The type of class, useful for recursion
    :return: A fully populated dictionary with field names converted to camelCase
    """
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    elif isinstance(obj, dict):
        data = {}
        for (k, v) in obj.items():
            # Runs a conversion of snake_case to camelCase to be compliant with the API
            data[underscore_to_camel(k)] = to_dict(v, cls)
        return data
    elif hasattr(obj, "_ast"):
        return to_dict(obj._ast())
    elif hasattr(obj, "__iter__") and not isinstance(obj, str):
        return [to_dict(v, cls) for v in obj]
    elif hasattr(obj, "__dict__"):
        data = dict([(underscore_to_camel(key), to_dict(value, cls))
                     for key, value in obj.__dict__.items()
                     if not callable(value) and not key.startswith('_')])
        if cls is not None and hasattr(obj, "__class__"):
            data[cls] = underscore_to_camel(obj.__class__.__name__)
        return data
    else:
        return obj


def from_json_string(s, object_hook):
    """
    Given the string as json, loads it into a nested dictionary and passes it into the object_ctor
    provided to yield an instance of the object
    :param s: A json formatted string
    :param object_hook: A function that takes a dictionary and yields a new object instance
    :return: A populated object instance generated from the object_ctor
    """
    d = json.loads(s)
    return object_hook(d)


def to_json_string(o):
    """
    Converts the object to json
    :param o: An object that can be serialized to json
    :return: A json formatted string representation of the object
    """
    return json.dumps(to_dict(o))


def map_option(v, f):
    """
    Applies the function f to the value if it is not None
    :param v: A value that maybe None
    :param f: A function that takes a single argument to apply to v
    :return: The result of applying f to v; None if v is None
    """
    if v:
        return f(v)
    else:
        return None


def parse_datetime(s):
    """
    Parses the iso formatted date from the string provided
    :param s: An iso date formatted string
    :return: A datetime
    """
    import dateutil.parser
    return dateutil.parser.parse(s)


def to_utc_strftime(t):
    """
    Converts the provided datetime to UTC then to a string
    Includes Python 2/3 compatibility
    :param t: Datetime with time zone
    :return: A string
    """
    if t.tzinfo is not None:
        try:
            return t.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
        except NameError:
            return t.astimezone(tzutc()).strftime("%Y-%m-%dT%H:%M:%SZ")
    else:
        raise ValueError("Expected datetime object with tzinfo attribute.")



# Copyright 2018 Comcast Cable Communications Management, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TODO: Add module docstring."""

import json
import logging
import os
from builtins import str

import requests
from future.moves.urllib.parse import parse_qs
from future.utils import iteritems
from requests.adapters import HTTPAdapter
# Python 2/3 compatibility
from requests.compat import urljoin
from requests.compat import urlparse
from requests.compat import urlsplit
from requests.packages.urllib3.util.retry import Retry

# TODO: Didn't like this boto request signer, fix when moving back
from vinyldns.boto_request_signer import BotoRequestSigner

from vinyldns.batch_change import BatchChange, ListBatchChangeSummaries, to_review_json
from vinyldns.membership import Group, ListGroupsResponse, ListGroupChangesResponse, ListMembersResponse, \
    ListAdminsResponse
from vinyldns.serdes import to_json_string
from vinyldns.zone import ListZonesResponse, ListZoneChangesResponse, Zone, ZoneChange
from vinyldns.record import ListRecordSetsResponse, ListRecordSetChangesResponse, RecordSet, RecordSetChange

try:
    basestring
except NameError:
    basestring = str

logger = logging.getLogger(__name__)

__all__ = [u'VinylDNSClient', u'MAX_RETRIES', u'RETRY_WAIT']

MAX_RETRIES = 30
RETRY_WAIT = 0.05


class ClientError(Exception):
    """Base class for custom exceptions"""
    pass


class BadRequestError(ClientError):
    """400 Bad Request error"""
    pass


class UnauthorizedError(ClientError):
    """401 Unauthorized error"""
    pass


class ForbiddenError(ClientError):
    """403 Forbidden Error"""
    pass


class NotFoundError(ClientError):
    """404 Not Found Error"""
    pass


class ConflictError(ClientError):
    """409 Conflict error"""
    pass


class UnprocessableError(ClientError):
    """422 Unprocessable Entity error"""
    pass


class VinylDNSClient(object):
    """TODO: Add class docstring."""

    def __init__(self, url, access_key, secret_key):
        """TODO: Add method docstring."""
        self.index_url = url
        self.headers = {
            u'Accept': u'application/json, text/plain',
            u'Content-Type': u'application/json'
        }

        self.signer = BotoRequestSigner(self.index_url,
                                        access_key, secret_key)

        self.session = self.__requests_retry_session()

    @classmethod
    def from_env(cls):
        """
        Create client from environment variables.

        :return: a client instance
        """
        url = os.environ.get('VINYLDNS_API_URL')
        access_key = os.environ.get('VINYLDNS_ACCESS_KEY_ID')
        secret_key = os.environ.get('VINYLDNS_SECRET_ACCESS_KEY')

        if url is None or access_key is None or secret_key is None:
            raise Exception('\'VINYLDNS_API_URL\', \'VINYLDNS_ACCESS_KEY_ID\', '
                            '\'VINYLDNS_SECRET_ACCESS_KEY\' environment variables'
                            'are required.')
        return cls(url, access_key, secret_key)

    def __requests_retry_session(self,
                                 retries=5,
                                 backoff_factor=0.4,
                                 status_forcelist=(500, 502, 504),
                                 session=None):

        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount(u'http://', adapter)
        session.mount(u'https://', adapter)
        return session

    def __make_request(self, url, method=u'GET', headers=None, body_string=None, **kwargs):

        # remove retries arg if provided
        kwargs.pop(u'retries', None)

        path = urlparse(url).path

        # we must parse the query string so we can provide it if it exists so that we can pass it to the
        # build_vinyldns_request so that it can be properly included in the AWS signing...
        query = parse_qs(urlsplit(url).query)

        if query:
            # the problem with parse_qs is that it will return a list for ALL params, even if they are a single value
            # we need to essentially flatten the params if a param has only one value
            query = dict((k, v if len(v) > 1 else v[0])
                         for k, v in iteritems(query))

        signed_headers, signed_body = self.__build_vinyldns_request(method, path, body_string, query,
                                                                    with_headers=headers or {}, **kwargs)

        response = self.session.request(method, url, data=signed_body, headers=signed_headers, **kwargs)

        return self.__check_response(response, method)

    def __check_response(self, response, method):
        status = response.status_code
        if status == 200 or status == 202:
            return response.status_code, response.json()
        elif status == 400:
            raise BadRequestError(response.text)
        elif status == 401:
            raise UnauthorizedError(response.text)
        elif status == 403:
            raise ForbiddenError(response.text)
        elif status == 404:
            if method == 'GET':
                return 404, None
            else:
                raise NotFoundError(response.text)
        elif status == 409:
            raise ConflictError(response.text)
        elif status == 422:
            raise UnprocessableError(response.text)
        else:
            raise ClientError(response.text)

    def __build_vinyldns_request(self, method, path, body_data, params=None, **kwargs):

        if isinstance(body_data, basestring):
            body_string = body_data
        else:
            body_string = json.dumps(body_data)

        new_headers = {u'X-Amz-Target': u'VinylDNS'}
        new_headers.update(kwargs.get(u'with_headers', dict()))

        suppress_headers = kwargs.get(u'suppress_headers', list())

        headers = self.__build_headers(new_headers, suppress_headers)

        auth_header = self.signer.build_auth_header(method, path, headers, body_string, params)
        headers[u'Authorization'] = auth_header

        return headers, body_string

    @staticmethod
    def __build_headers(new_headers, suppressed_keys):

        def canonical_header_name(field_name):
            return u'-'.join(word.capitalize() for word in field_name.split(u'-'))

        import datetime
        now = datetime.datetime.utcnow()
        headers = {u'Content-Type': u'application/x-amz-json-1.0',
                   u'Date': now.strftime(u'%a, %d %b %Y %H:%M:%S GMT'),
                   u'X-Amz-Date': now.strftime(u'%Y%m%dT%H%M%SZ')}

        for k, v in iteritems(new_headers):
            headers[canonical_header_name(k)] = v

        for k in map(canonical_header_name, suppressed_keys):
            if k in headers:
                del headers[k]

        return headers

    def create_group(self, group, **kwargs):
        """
        Create a new group.

        :param group: A group dictionary that can be serialized to json
        :return: the content of the response, which should be a group json
        """
        url = urljoin(self.index_url, u'/groups')
        response, data = self.__make_request(url, u'POST', self.headers, to_json_string(group), **kwargs)

        return Group.from_dict(data)

    def get_group(self, group_id, **kwargs):
        """
        Get a group.

        :param group_id: Id of the group to get
        :return: the group json
        """
        url = urljoin(self.index_url, u'/groups/' + group_id)
        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)

        return Group.from_dict(data) if data is not None else None

    def delete_group(self, group_id, **kwargs):
        """
        Delete a group.

        :param group_id: Id of the group to delete
        :return: the group json
        """
        url = urljoin(self.index_url, u'/groups/' + group_id)
        response, data = self.__make_request(url, u'DELETE', self.headers, **kwargs)

        return Group.from_dict(data)

    def update_group(self, group, **kwargs):
        """
        Update an existing group, uses the id of the group provided

        :param group: A group to be updated
        :return: the content of the response, which should be a group json
        """
        url = urljoin(self.index_url, u'/groups/{0}'.format(group.id))
        response, data = self.__make_request(url, u'PUT', self.headers, to_json_string(group), **kwargs)

        return Group.from_dict(data)

    def list_my_groups(self, group_name_filter=None, start_from=None, max_items=None, **kwargs):
        """
        Retrieve my groups.

        :param start_from: the start key of the page; this is the next_id of a prior call
        :param max_items: the number of groups to return
        :param group_name_filter: only returns groups whose names contain filter string
        :return: the content of the response
        """
        args = []
        if group_name_filter:
            args.append(u'groupNameFilter={0}'.format(group_name_filter))
        if start_from:
            args.append(u'startFrom={0}'.format(start_from))
        if max_items is not None:
            args.append(u'maxItems={0}'.format(max_items))

        url = urljoin(self.index_url, u'/groups') + u'?' + u'&'.join(args)
        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)

        return ListGroupsResponse.from_dict(data)

    def list_all_my_groups(self, group_name_filter=None, **kwargs):
        """
        Retrieve all my groups, paging through the results until exhausted

        :param group_name_filter: only returns groups whose names contain filter string
        :return: the content of the response
        """
        groups = []
        args = []
        if group_name_filter:
            args.append(u'groupNameFilter={0}'.format(group_name_filter))

        url = urljoin(self.index_url, u'/groups') + u'?' + u'&'.join(args)
        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)
        groups.extend(data[u'groups'])

        while u'nextId' in data and data[u'nextId']:
            next_args = args[:]
            next_args.append(u'startFrom={0}'.format(data['nextId']))
            url = urljoin(self.index_url, u'/groups') + u'?' + u'&'.join(next_args)
            response, data = self.__make_request(url, u'GET', self.headers, **kwargs)
            groups.extend(data[u'groups'])

        g = [Group.from_dict(elem) for elem in groups]
        return ListGroupsResponse(groups=g, group_name_filter=group_name_filter)

    def list_members_group(self, group_id, start_from=None, max_items=None, **kwargs):
        """
        List the members of an existing group.

        :param group_id: the Id of an existing group
        :param start_from: the Id a member of the group
        :param max_items: the max number of items to be returned
        :return: the json of the members
        """
        if start_from is None and max_items is None:
            url = urljoin(self.index_url, u'/groups/{0}/members'.format(group_id))
        elif start_from is None and max_items is not None:
            url = urljoin(self.index_url, u'/groups/{0}/members?maxItems={1}'.format(group_id, max_items))
        elif start_from is not None and max_items is None:
            url = urljoin(self.index_url, u'/groups/{0}/members?startFrom={1}'.format(group_id, start_from))
        elif start_from is not None and max_items is not None:
            url = urljoin(self.index_url, u'/groups/{0}/members?startFrom={1}&maxItems={2}'.format(group_id,
                                                                                                   start_from,
                                                                                                   max_items))

        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)

        return ListMembersResponse.from_dict(data)

    def list_group_admins(self, group_id, **kwargs):
        """
        Return the group admins.

        :param group_id: the Id of the group
        :return: the user info of the admins
        """
        url = urljoin(self.index_url, u'/groups/{0}/admins'.format(group_id))
        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)

        return ListAdminsResponse.from_dict(data)

    def list_group_changes(self, group_id, start_from=None, max_items=None, **kwargs):
        """
        List the changes of an existing group.

        :param group_id: the Id of an existing group
        :param start_from: the Id a group change
        :param max_items: the max number of items to be returned
        :return: the json of the members
        """
        if start_from is None and max_items is None:
            url = urljoin(self.index_url, u'/groups/{0}/activity'.format(group_id))
        elif start_from is None and max_items is not None:
            url = urljoin(self.index_url, u'/groups/{0}/activity?maxItems={1}'.format(group_id, max_items))
        elif start_from is not None and max_items is None:
            url = urljoin(self.index_url, u'/groups/{0}/activity?startFrom={1}'.format(group_id, start_from))
        elif start_from is not None and max_items is not None:
            url = urljoin(self.index_url, u'/groups/{0}/activity?startFrom={1}&maxItems={2}'.format(group_id,
                                                                                                    start_from,
                                                                                                    max_items))

        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)

        return ListGroupChangesResponse.from_dict(data)

    def connect_zone(self, zone, **kwargs):
        """
        Create a new zone with the given name and email.

        :param zone: the zone to be created
        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones')
        response, data = self.__make_request(url, u'POST', self.headers, to_json_string(zone), **kwargs)
        return ZoneChange.from_dict(data)

    def update_zone(self, zone, **kwargs):
        """
        Update a zone.

        :param zone: the zone to be created
        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones/{0}'.format(zone.id))
        response, data = self.__make_request(url, u'PUT', self.headers, to_json_string(zone), **kwargs)
        return ZoneChange.from_dict(data)

    def sync_zone(self, zone_id, **kwargs):
        """
        Sync a zone.

        :param zone: the zone to be updated
        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones/{0}/sync'.format(zone_id))
        response, data = self.__make_request(url, u'POST', self.headers, **kwargs)

        return ZoneChange.from_dict(data)

    def abandon_zone(self, zone_id, **kwargs):
        """
        Delete the zone for the given id.

        :param zone_id: the id of the zone to be deleted
        :return: nothing, will fail if the status code was not expected
        """
        url = urljoin(self.index_url, u'/zones/{0}'.format(zone_id))
        response, data = self.__make_request(url, u'DELETE', self.headers, **kwargs)

        return ZoneChange.from_dict(data)

    def get_zone(self, zone_id, **kwargs):
        """
        Get a zone for the given zone id.

        :param zone_id: the id of the zone to retrieve
        :return: the zone, or will 404 if not found
        """
        url = urljoin(self.index_url, u'/zones/{0}'.format(zone_id))
        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)

        return Zone.from_dict(data['zone']) if data is not None else None

    def get_zone_by_name(self, name, **kwargs):
        """
        Get a zone by zone name.
        :param zone: the name of the zone to retrieve
        :return: the zone, or will 404 if not found
        """
        url = urljoin(self.index_url, u'/zones/name/{0}'.format(name))
        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)
        return Zone.from_dict(data['zone']) if data is not None else None

    def list_zone_changes(self, zone_id, start_from=None, max_items=None, **kwargs):
        """
        Get the zone changes for the given zone id.

        :param zone_id: the id of the zone to retrieve
        :param start_from: the start key of the page
        :param max_items: the page limit
        :return: the zone, or will 404 if not found
        """
        args = []
        if start_from:
            args.append(u'startFrom={0}'.format(start_from))
        if max_items is not None:
            args.append(u'maxItems={0}'.format(max_items))
        url = urljoin(self.index_url, u'/zones/{0}/changes'.format(zone_id)) + u'?' + u'&'.join(args)

        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)
        return ListZoneChangesResponse.from_dict(data)

    def list_zones(self, name_filter=None, start_from=None, max_items=None, **kwargs):
        """
        Get a list of zones that currently exist.

        :return: a list of zones
        """
        url = urljoin(self.index_url, u'/zones')

        query = []
        if name_filter:
            query.append(u'nameFilter=' + name_filter)

        if start_from:
            query.append(u'startFrom=' + str(start_from))

        if max_items:
            query.append(u'maxItems=' + str(max_items))

        if query:
            url = url + u'?' + u'&'.join(query)

        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)
        return ListZonesResponse.from_dict(data)

    def create_record_set(self, record_set, **kwargs):
        """
        Create a new record_set.

        :param record_set: the record_set to be created
        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones/{0}/recordsets'.format(record_set.zone_id))
        response, data = self.__make_request(url, u'POST', self.headers, to_json_string(record_set), **kwargs)
        return RecordSetChange.from_dict(data)

    def delete_record_set(self, zone_id, rs_id, **kwargs):
        """
        Delete an existing record_set.

        :param zone_id: the zone id the record_set belongs to
        :param rs_id: the id of the record_set to be deleted
        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones/{0}/recordsets/{1}'.format(zone_id, rs_id))

        response, data = self.__make_request(url, u'DELETE', self.headers, **kwargs)
        return RecordSetChange.from_dict(data)

    def update_record_set(self, record_set, **kwargs):
        """
        Delete an existing record_set.

        :param record_set: the record_set to be updated
        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones/{0}/recordsets/{1}'.format(record_set.zone_id, record_set.id))

        response, data = self.__make_request(url, u'PUT', self.headers,
                                             to_json_string(record_set), **kwargs)

        return RecordSetChange.from_dict(data)

    def get_record_set(self, zone_id, rs_id, **kwargs):
        """
        Get an existing record_set.

        :param zone_id: the zone id the record_set belongs to
        :param rs_id: the id of the record_set to be retrieved
        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones/{0}/recordsets/{1}'.format(zone_id, rs_id))

        response, data = self.__make_request(url, u'GET', self.headers, None, **kwargs)
        return RecordSet.from_dict(data['recordSet']) if data is not None else None

    def list_record_sets(self, zone_id, start_from=None, max_items=None, record_name_filter=None, **kwargs):
        """
        Retrieve record_sets in a zone.

        :param zone_id: the zone to retrieve
        :param start_from: the start key of the page
        :param max_items: the page limit
        :param record_name_filter: only returns record_sets whose names contain filter string
        :return: the content of the response
        """
        args = []
        if start_from:
            args.append(u'startFrom={0}'.format(start_from))
        if max_items is not None:
            args.append(u'maxItems={0}'.format(max_items))
        if record_name_filter:
            args.append(u'recordNameFilter={0}'.format(record_name_filter))

        url = urljoin(self.index_url, u'/zones/{0}/recordsets'.format(zone_id)) + u'?' + u'&'.join(args)

        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)
        return ListRecordSetsResponse.from_dict(data)

    def get_record_set_change(self, zone_id, rs_id, change_id, **kwargs):
        """
        Get an existing record_set change.

        :param zone_id: the zone id the record_set belongs to
        :param rs_id: the id of the record_set to be retrieved
        :param change_id: the id of the change to be retrieved
        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones/{0}/recordsets/{1}/changes/{2}'.format(zone_id, rs_id, change_id))

        response, data = self.__make_request(url, u'GET', self.headers, None, **kwargs)
        return RecordSetChange.from_dict(data) if data is not None else None

    def list_record_set_changes(self, zone_id, start_from=None, max_items=None, **kwargs):
        """
        Get the record_set changes for the given zone id.

        :param zone_id: the id of the zone to retrieve
        :param start_from: the start key of the page
        :param max_items: the page limit
        :return: the zone, or will 404 if not found
        """
        args = []
        if start_from:
            args.append(u'startFrom={0}'.format(start_from))
        if max_items is not None:
            args.append(u'maxItems={0}'.format(max_items))
        url = urljoin(self.index_url, u'/zones/{0}/recordsetchanges'.format(zone_id)) + u'?' + u'&'.join(args)

        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)
        return ListRecordSetChangesResponse.from_dict(data)

    def create_batch_change(self, batch_change_input, allow_manual_review=None, **kwargs):
        """
        Create a new batch change.

        :param batch_change_input: the batchchange to be created
        :param allow_manual_review: set to false to fail rather than go to
        review if there are errors
        :return: the content of the response
        """
        arg = ''

        if allow_manual_review is not None:
            arg = u'allowManualReview={0}'.format(allow_manual_review)

        url = urljoin(self.index_url, u'/zones/batchrecordchanges') + u'?' + arg
        response, data = self.__make_request(url, u'POST', self.headers, to_json_string(batch_change_input), **kwargs)

        return BatchChange.from_dict(data)

    def get_batch_change(self, batch_change_id, **kwargs):
        """
        Get an existing batch change.

        :param batch_change_id: the unique identifier of the batchchange
        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones/batchrecordchanges/{0}'.format(batch_change_id))
        response, data = self.__make_request(url, u'GET', self.headers, None, **kwargs)

        return BatchChange.from_dict(data) if data is not None else None

    def list_batch_change_summaries(self, start_from=None, max_items=None,
                                    ignore_access=None, approval_status=None, **kwargs):
        """
        Get list of user's batch change summaries.

        :return: the content of the response
        """
        args = []
        if start_from:
            args.append(u'startFrom={0}'.format(start_from))
        if max_items is not None:
            args.append(u'maxItems={0}'.format(max_items))
        if ignore_access:
            args.append(u'ignoreAccess={0}'.format(ignore_access))
        if approval_status:
            args.append(u'approvalStatus={0}'.format(approval_status))

        url = urljoin(self.index_url, u'/zones/batchrecordchanges') + u'?' + u'&'.join(args)

        response, data = self.__make_request(url, u'GET', self.headers, **kwargs)
        return ListBatchChangeSummaries.from_dict(data)

    def approve_batch_change(self, batch_change_id, approval=None, **kwargs):
        """
        Approve a batch change

        :return: the content of the response
        """
        url = urljoin(self.index_url, '/zones/batchrecordchanges/{0}/approve'.format(batch_change_id),
                      to_json_string(approval))
        response, data = self.__make_request(url, u'POST', self.headers, to_review_json(approval), **kwargs)

        return BatchChange.from_dict(data)

    def cancel_batch_change(self, batch_change_id, **kwargs):
        """
        Cancel a batch change

        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones/batchrecordchanges/{0}/cancel'.format(batch_change_id))
        response, data = self.__make_request(url, u'POST', self.headers, **kwargs)

        return BatchChange.from_dict(data) if data is not None else None

    def reject_batch_change(self, batch_change_id, rejection=None, **kwargs):
        """
        Reject a batch change

        :return: the content of the response
        """
        url = urljoin(self.index_url, u'/zones/batchrecordchanges/{0}/reject'.format(batch_change_id))
        response, data = self.__make_request(url, u'POST', self.headers,  to_review_json(rejection), **kwargs)

        return BatchChange.from_dict(data)

    def add_zone_acl_rule(self, zone_id, acl_rule, **kwargs):
        """
        Put an acl rule on the zone.

        :param zone_id: The id of the zone to attach the acl rule to
        :param acl_rule: The acl rule contents
        :return: the content of the response
        """
        url = urljoin(self.index_url, '/zones/{0}/acl/rules'.format(zone_id))
        response, data = self.__make_request(url, 'PUT', self.headers,
                                             to_json_string(acl_rule), **kwargs)

        return ZoneChange.from_dict(data)

    def delete_zone_acl_rule(self, zone_id, acl_rule, **kwargs):
        """
        Delete an acl rule from the zone.

        :param zone_id: The id of the zone to remove the acl from
        :param acl_rule: The acl rule to remove
        :return: the content of the response
        """
        url = urljoin(self.index_url, '/zones/{0}/acl/rules'.format(zone_id))
        response, data = self.__make_request(url, 'DELETE', self.headers,
                                             to_json_string(acl_rule), **kwargs)

        return ZoneChange.from_dict(data)



# Copyright 2018 Comcast Cable Communications Management, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function

import sys
from glob import glob
from os.path import basename
from os.path import splitext

from setuptools import find_packages
from setuptools import setup
from setuptools.command.test import test as TestCommand


# Allows one to simply > python3 setup.py test
class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


with open('README.md') as f:
    long_description = f.read()

setup(
    cmdclass={'test': PyTest},
    name='vinyldns-python',
    version='0.9.6',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    url='https://github.com/vinyldns/vinyldns-python',
    license='Apache Software License 2.0',
    author='vinyldns',
    author_email='vinyldns-core@googlegroups.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Internet :: Name Service (DNS)',
        'License :: OSI Approved :: Apache Software License',
    ],
    keywords=[
        'dns', 'python', 'vinyldns',
    ],
    description='Python client library for VinylDNS',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        'boto>=2.48.0',
        'future>=0.17.1',
        'requests>=2.20.0',
        'python-dateutil>=2.7.5',
    ],
    tests_require=[
        'responses==0.10.4',
        'pytest==3.10.1',
    ],
)



# Copyright 2018 Comcast Cable Communications Management, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TODO: Add module docstring."""
from vinyldns.serdes import parse_datetime, map_option


class AccessLevel:
    NoAccess = "NoAccess"
    Read = "Read"
    Write = "Write"
    Delete = "Delete"


class ACLRule(object):
    def __init__(self, access_level, description=None, user_id=None, group_id=None, record_mask=None, record_types=[]):
        self.access_level = access_level
        self.description = description
        self.user_id = user_id
        self.group_id = group_id
        self.record_mask = record_mask
        self.record_types = record_types

    @staticmethod
    def from_dict(d):
        return ACLRule(access_level=d['accessLevel'], description=d.get('description'),
                       user_id=d.get('userId'), group_id=d.get('groupId'), record_mask=d.get('recordMask'),
                       record_types=d.get('recordTypes', []))


class ZoneACL(object):
    def __init__(self, rules=[]):
        self.rules = rules

    @staticmethod
    def from_dict(d):
        rules = [ACLRule.from_dict(r) for r in d.get('rules', [])]
        return ZoneACL(rules)


class ZoneConnection(object):
    def __init__(self, name, key_name, key, primary_server):
        self.name = name
        self.key_name = key_name
        self.key = key
        self.primary_server = primary_server

    @staticmethod
    def from_dict(d):
        return ZoneConnection(name=d['name'], key_name=d['keyName'], key=d['key'],
                              primary_server=d['primaryServer'])


class Zone(object):
    def __init__(self, name, email, admin_group_id, id=None, status=None, created=None, updated=None, connection=None,
                 transfer_connection=None, acl=ZoneACL(), latest_sync=None, is_test=False, shared=False,
                 backend_id=None):
        self.id = id
        self.name = name
        self.email = email
        self.status = status
        self.created = created
        self.updated = updated
        self.connection = connection
        self.transfer_connection = transfer_connection
        self.acl = acl
        self.admin_group_id = admin_group_id
        self.latest_sync = latest_sync
        self.is_test = is_test
        self.shared = shared
        self.backend_id = backend_id

    @staticmethod
    def from_dict(d):
        """
        from_dicts the string formatted json into a zone
        :param d: A dictionary built from json
        :return: A populated zone
        """
        return Zone(
            id=d['id'],
            name=d['name'],
            email=d['email'],
            admin_group_id=d['adminGroupId'],
            status=d.get('status'),
            created=map_option(d.get('created'), parse_datetime),
            updated=map_option(d.get('updated'), parse_datetime),
            connection=map_option(d.get('connection'), ZoneConnection.from_dict),
            transfer_connection=map_option(d.get('transferConnection'), ZoneConnection.from_dict),
            acl=map_option(d.get('acl'), ZoneACL.from_dict),
            latest_sync=map_option(d.get('latestSync'), parse_datetime),
            is_test=d.get('isTest', False),
            shared=d['shared'],
            backend_id=d.get('backendId')
        )


class ListZonesResponse(object):
    def __init__(self, zones, name_filter, start_from=None, next_id=None, max_items=100):
        self.zones = zones
        self.name_filter = name_filter
        self.start_from = start_from
        self.next_id = next_id
        self.max_items = max_items

    @staticmethod
    def from_dict(d):
        zones = [Zone.from_dict(elem) for elem in d.get('zones', [])]
        return ListZonesResponse(zones=zones, name_filter=d.get('nameFilter'), start_from=d.get('startFrom'),
                                 next_id=d.get('nextId'), max_items=d.get('maxItems', 100))


class ZoneChange(object):
    def __init__(self, zone, user_id, change_type, status, created, system_message, id):
        self.zone = zone
        self.user_id = user_id
        self.change_type = change_type
        self.status = status
        self.created = created
        self.system_message = system_message
        self.id = id

    @staticmethod
    def from_dict(d):
        zone = Zone.from_dict(d['zone'])
        created = map_option(d.get('created'), parse_datetime)
        return ZoneChange(zone=zone, user_id=d['userId'], change_type=d['changeType'], status=d['status'],
                          created=created, system_message=d.get('systemMessage'), id=d['id'])


class ListZoneChangesResponse(object):
    def __init__(self, zone_id, zone_changes, start_from=None, next_id=None, max_items=100):
        self.zone_id = zone_id
        self.zone_changes = zone_changes
        self.next_id = next_id
        self.start_from = start_from
        self.max_items = max_items

    @staticmethod
    def from_dict(d):
        zone_changes = [ZoneChange.from_dict(elem) for elem in d.get('zoneChanges', [])]
        return ListZoneChangesResponse(zone_id=d['zoneId'], zone_changes=zone_changes, next_id=d.get('nextId'),
                                       start_from=d.get('startFrom'), max_items=d.get('maxItems', 100))
